#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <windows.h>
#include "GetArguments.h"
#include "../ReferencesFileInfo.h"

// Required is target file
#define MAX_LINE_LENGTH 1024
#define MAX_MESSAGE_LENGTH 255

HRESULT result;
referencesFileHeader references_header_info = 
{
	.max_line_length = 0,
	.number_debug_messages = 0,
	.version_major = 0,
	.version_minor = 1
};

HRESULT SetupFiles(char * filepath, FILE ** input_file, FILE ** temporary_file);
bool ReadNextLine(FILE * input_file, FILE * temporary_file, FILE * output_file);
HRESULT OpenNextFile(char * filepath, FILE ** input_file, FILE ** temporary_file);
HRESULT OpenInputFile(char * input_file_path, FILE ** input_file, FILE ** temporary_file);
void DebugTokenLine(FILE * temporary_file, FILE * output_file, char * line_copy);
void PrintHeaderInformation(FILE * file,  fpos_t header_start);
void PadFile(FILE * file);
void GetTempFilePath(char * input_file_path, char ** temp_file_path);


int main (int argc, char ** argv)
{

	if (argc > 255)
	{
		printf("Too many arguments\n");
		return 0;
	}
	if(!CheckArguments(&application_arguments, argc, argv))
	{
		printf("No input files given\n");
		return 1;
	}
	//PrintInputFilePaths(&application_arguments, argv);
	
	// This is the file to be read from
	FILE * input_file = NULL;
	/**
	 * @brief Copy of input file with debug tokens changed.
	 * 
	 * @note This is created for each input file and is essentially a copy of the file.
	 * 		However, it has the debug tokens replaced with the desired function call.
	 * 		Each debug token has a unique value which is correlated in the references file.
	 *		This temporary file will have the same name but a .temp extension appended
	 */
	FILE * temporary_file = NULL;
	/**
	 * @brief Output file for references
	 * 
	 * @note This is the reference file which will be generated and is to be passed to the receiver program at runtime.
	 * 		This will create a new file on each run so care should be taken to ensure that it doesn't overwrite an older file.
	 */
	FILE * output_file = fopen(application_arguments.references_output_file_location, "w");
	
	// This will be needed later to write the header to the file
	fpos_t output_file_start;
	fgetpos(output_file, &output_file_start);
	PadFile(output_file);
	
	// Go through and process each file from the list
	while (1)
	{
		static uint8_t i = 0;
		// If argument_index is 0 then the list of arguments is finished 
		uint8_t argument_index = *(application_arguments.args_list_input_files_indexs + i);
		if (!argument_index)
		{
			break;
		}

		char * next_file_name = *(argv + argument_index); 
		
		if (!OpenInputFile(next_file_name, &input_file, &temporary_file))
		{
			// Read and process all the lines out of the file
			while(ReadNextLine(input_file, temporary_file, output_file));
			
			// Close the input file and its temporary copy
			fclose(input_file);
			fclose(temporary_file);
		}

		i++;
	}
	
	// Finished reading through all the file and editing, now put the header with metadata at the top of the output file
	PrintHeaderInformation(output_file, output_file_start);
	fclose(output_file);
	printf("Done\n");

	return S_OK;
}


/**
 * @brief Reads the next line of the input file and decides if it has an input token. If it does then sends it to the debug token handler function.
 * 
 * @param[in] input_file File to read
 * @param[in] temporary_file File to copy out either unedited lines or edited lines with debug token substituted 
 * @param[in] output_file File to put associations of debug message IDs and the messages
 * 
 * @return true : Another line available to be read in file.
 * @return false : EOF.
 */
bool ReadNextLine(FILE * input_file, FILE * temporary_file, FILE * output_file)
{

	static uint16_t line_counter = 0;
	char line[MAX_LINE_LENGTH] = {0};
	// fgets() returns null for EOF
	if(!fgets(line, sizeof(line), input_file))
	{
		//printf("End of file\n");
		line_counter = 0;
		return false;
	}
	
	char line_copy[MAX_LINE_LENGTH];
	// Necessary as strtok() edits the base string
	strcpy(line_copy, line);

	// Find out if there is a debug token on this line
	// Ignore lines with the debug token which are just for #defines to keep the compiler/IDE happy
	const char define_token[] = "#define";
	const char token_delimiters[] = "\t\n\r \";()";

	uint8_t token_counter = 0;
	char * token = strtok(line, token_delimiters);
	bool debug_token_on_line = false;
	
	while(token)
	{
		//printf("%d.%d : %s\n", line_counter, token_counter, token);
		if (!strcmp(define_token, token))
		{
			// Token is on line as part of a define, don't include it
			break;
		} else if (!strcmp(application_arguments.debug_token, token))
		{
			// Debug token found
			debug_token_on_line = true;
		}

		token_counter++;
		token = strtok(NULL, token_delimiters);			
	}

	if (debug_token_on_line)
	{
		DebugTokenLine(temporary_file, output_file, line_copy);
	} else 
	{
		fputs(line_copy, temporary_file);
	}
		
	line_counter++;
	return true;
}


/**
 * @brief Takes a line with a debug token gives it an ID and prints out the edited line to the temp file and references file 
 * 
 * @param[in] temporary_file Pointer to FILE handler for the temporary file
 * @param[in] output_file Pointer to FILE handler for the output references file
 * @param[in] debug_line LSine containing the debug token
 */
void DebugTokenLine(FILE * temporary_file, FILE * output_file, char * debug_line)
{
	// Read the tabs/whitespace before the start of the debug token
	char * tabs_token = strtok(debug_line, application_arguments.debug_token);
	// Read and then ignore up to the first double quotes, this will be the debug token
	strtok(NULL, "\"");
	// Read the debug message into the token 
	char * message_token = strtok(NULL, "\"");

	// Keep the max token length updated in the references header information struct
	uint16_t token_length = strlen(message_token);
	if ((token_length + 1) > references_header_info.max_line_length)
	{
		// Needs to be +2 to fit whitespace and newline character of the largest message
		references_header_info.max_line_length = token_length + 2;
	}
	
	// Create the new line to go into the temp file. Also put in unique value for the number of the debug message
	fprintf(temporary_file, "%s%s(%hhu);\n", 
				tabs_token, 
				application_arguments.replacement_token, 
				references_header_info.number_debug_messages + 1); // Needs to be + 1 to avoid zero, null character can't be easiy received

	// This uses a formatted string to allow for printing the integer value correctly as a character in the text file
	fprintf(output_file, "%hhu : %s\n", references_header_info.number_debug_messages + 1, message_token);

	// Finally, increment the debug message counter
	references_header_info.number_debug_messages++;
	
	return;
}


/**
 * @brief Prints the header information into the file at the position given by header_start
 * 
 * @remark This is intended to be used at the end of the program to print the header information
 * 			at the top of the references file. It will overwrite anything already at header_start and 
 * 			so padding should be added at header_start when the file is first created.
 * 
 * @param[in] file File to write the header in 
 * @param[in] header_start Position in file to start writing header  
 */
void PrintHeaderInformation(FILE * file, fpos_t header_start)
{
	// Header information is held globally
	fsetpos(file, &header_start);
	// Newline is handled by the padding
	fprintf(file, "Version : %hhu.%hhu, Debug messages : %hhu, Max line length : %hu",
						references_header_info.version_major,
						references_header_info.version_minor,
						references_header_info.number_debug_messages,
						references_header_info.max_line_length);
	
	return;
}


/**
 * @brief Adds a lump of whitespace terminated with a newline to the start of the file 
 * 
 * @note This needs to be done so the header can be added without overwriting data at the end.
 * 		It's a bit of a bodge and relys on the header being no larger than a hard-coded size
 * 		At this point the header structure is 
 * 		"Version : %hhu.%hhu, Debug messages : %hhu, Max line length : %hu\n"
 * 		Which gives it a maximum length of <70 characters (64 I believe)
 * 		So the file start will be padded with empty characters
 * 		There can be extra whitespace without any issues
 * 
 * @param file Output references file to add the padding to 
 */
void PadFile(FILE * file)
{
	const char newline = '\n';
	// Padding length has to be hardcoded into printf
	fprintf(file, "%70c", newline);
	return;
}


/**
 * @brief Opens the file given by the input path creates a temporary file to store the edited file
 * 
 * @param[in] input_file_path Path of input file
 * @param[out] input_file Address of pointer to file handler for input file 
 * @param[out] temporary_file Address of pointer to file handler for output file
 *
 * @return Result of trying to open files 
 */
HRESULT OpenInputFile(char * input_file_path, FILE ** input_file, FILE ** temporary_file)
{
	printf("Copying and reading from %s\n", input_file_path);
	
	// Open input file for read fopen returns null if input file doesn't exist.
	*input_file = fopen(input_file_path, "r");
	if (!*input_file)
	{
		printf("File not found : %s\n", input_file_path);
		return HFILE_ERROR;
	}


	// To Do: This should leave the extension as .c or .cpp, instead ".temp" should be added 
	// between the name and extension, i.e. "main.temp.c"
	char * temp_file_path;
	GetTempFilePath(input_file_path, &temp_file_path);
	
	*temporary_file = fopen(temp_file_path, "w");
	free(temp_file_path);
	
	return S_OK;
}


/**
 * @brief Get a filepath for the temp file. This is a bit of janky way to do it.
 * 
 * @param[in] input_file_path 
 * @param[out] temp_file_path 
 */
void GetTempFilePath(char * input_file_path, char ** temp_file_path)
{
	// Length of the filename extension, i.e. ".c" or ".cpp". Start from one as there is always '\n'.
	uint8_t extension_length = 1;
	const uint8_t input_file_path_length = strlen(input_file_path);

	while(1)
	{
		// Go from the end of the path backwards until the first period seperator
		if(*(input_file_path + (input_file_path_length - extension_length)) == '.')
		{
			// This will count the newline character as the extension
			extension_length++;
			break;
		}
		extension_length++;
	}

	// Now need to create the name with the temp extension
	const uint8_t temp_token_length = strlen(application_arguments.temp_file_token);
	*temp_file_path = (char *) malloc(sizeof(char) * (input_file_path_length + temp_token_length) + 1);

	// Places the temp string from the app arguments between the name and the extension
	const uint8_t path_length_without_extension = (input_file_path_length - extension_length) + 1;
	memcpy(*temp_file_path, input_file_path, path_length_without_extension);
	memcpy((*temp_file_path) + path_length_without_extension, 
			application_arguments.temp_file_token, 
			temp_token_length);
	memcpy((*temp_file_path) + (path_length_without_extension + temp_token_length), 
			(input_file_path + path_length_without_extension),
			(extension_length + 1)); // + 1 to include null termination
	return;
}