#include "SerialComms.h"

/**
 * @brief Opens a serial port at the desired location and baudrate.
 * 
 * @param[out] com_handle_pointer Pointer to variable for handler to be written to 
 * @param[in] com_name String name of COM port, i.e. "COM1"
 * @param[in] baudrate Unsigned integer baudrate for port 
 * @return Result code
 */
serialReturnCode OpenSerialPort(HANDLE * com_handle_pointer, char * com_name, unsigned long baudrate)
{
    // printf("Trying to open serial port.\n");
    // https://learn.microsoft.com/en-us/windows/win32/devio/configuring-a-communications-resource
    BOOL   status;
    DCB dcb_serial_params = { 0 };  // Initializing DCB structure, must be zeroed
    COMMTIMEOUTS timeouts = { 0 };  //Initializing timeouts structure
    HANDLE com_handle = *com_handle_pointer;

    // Open the serial com port
    // This needs to be appended to the serial port name for it to be found
    const char * com_name_prefix = "\\\\.\\";
    uint8_t length_com_name = strlen(com_name) + sizeof(com_name_prefix) + 1;
    char * appended_com_name = (char *) malloc(length_com_name);
    // Bit annoying because sizeof() includes the escaping backslash characters in the length
    memcpy(appended_com_name, com_name_prefix, sizeof(com_name_prefix) - 4);
    memcpy((appended_com_name + sizeof(com_name_prefix) - 4), com_name, length_com_name - sizeof(com_name_prefix));

    com_handle = CreateFile(appended_com_name, //friendly name
                       GENERIC_READ | GENERIC_WRITE,      // Read/Write Access
                       0,                                 // No Sharing, ports cant be shared
                       NULL,                              // Default Security
                       OPEN_EXISTING,                     // Open existing port only
                       0,                                 // Non Overlapped I/O
                       NULL);                             // Must be NULL for COM devices


    free(appended_com_name);

	if (com_handle == INVALID_HANDLE_VALUE)
    {
        printf_s("\n Port can't be opened, error %d\n\n", GetLastError());
        return SERIAL_PORT_NOT_OPENED;
    }

    printf_s("Port opened on %s\n", com_name);

    //Setting the Parameters for the SerialPort
    dcb_serial_params.DCBlength = sizeof(dcb_serial_params);

    status = GetCommState(com_handle, &dcb_serial_params); // Retrieves the current settings
    if (status == FALSE)
    {
        printf_s("\nError to Get the Com state\n\n");
        return SERIAL_STATE_RETRIEVAL_ERROR;
    }
    // printf_s("Got COM state.\n");

    dcb_serial_params.BaudRate = baudrate;
    dcb_serial_params.ByteSize = 8;
    dcb_serial_params.StopBits = ONESTOPBIT;
    dcb_serial_params.Parity = NOPARITY;

    status = SetCommState(com_handle, &dcb_serial_params);
    if (status == FALSE)
    {
        printf_s("\nError to Setting DCB Structure\n\n");
        return SERIAL_DCB_ERROR;
    }

    //Setting Timeouts
    timeouts.ReadIntervalTimeout = 50;
    timeouts.ReadTotalTimeoutConstant = 50;
    timeouts.ReadTotalTimeoutMultiplier = 10;
    timeouts.WriteTotalTimeoutConstant = 50;
    timeouts.WriteTotalTimeoutMultiplier = 10;

    if (SetCommTimeouts(com_handle, &timeouts) == FALSE)
    {
        printf_s("\nError to Setting Time outs");
        return 1;
    }  

    *com_handle_pointer = com_handle;

	return S_OK;
}

/**
 * @brief Prints a string to a serial COM port.
 * 
 * @param[in] com_handle_pointer Pointer to file handler for COM port
 * @param[in] message Character string to send
 * @param[in] newline Boolean for whether a newline is to be appended
 * @return serialReturnCode 
 */
serialReturnCode SendOverSerial(HANDLE * com_handle_pointer, char * message, BOOL newline)
{
	HANDLE com_handle = *com_handle_pointer;
    DWORD BytesWritten;
	//Writing data to Serial Port
    BOOL status = WriteFile(com_handle,// Handle to the Serialport
                       message,            // Data to be written to the port
                       strlen(message),   // No of bytes to write into the port
                       &BytesWritten,  // No of bytes written to the port
                       NULL);
                       
    if (status == FALSE)
    {
        printf_s("\nFail to Write to COM port");
        return 1;
    }

    printf("%lu bytes written to COM port\n", BytesWritten);
	
	if (newline)
	{
		status = WriteFile(com_handle,// Handle to the Serialport
                       "\n",            // Data to be written to the port
                       1,   // No of bytes to write into the port
                       &BytesWritten,  // No of bytes written to the port
                       NULL);

	}

	if (status == FALSE)
    {
        printf_s("\nFail to Write to COM port");
        return 1;
    }

	return S_OK;
}

/**
 * @brief Reads the specified number of bytes from an already opened serial port into a message buffer.
 * 
 * @param[in] com_handle_pointer Pointer to file handle for the serial port
 * @param[out] message Buffer to read serial data into
 * @param[in] buffer_size Size of buffer in bytes
 * @return serialReturnCode 
 */
serialReturnCode ReceiveOverSerial(HANDLE * com_handle_pointer, char * message, uint8_t buffer_size)
{
    serialReturnCode return_status = SERIAL_RETURN_OK;
    // Leave room to append a null character
    // const uint8_t actual_read_size = buffer_size - 1;
    HANDLE com_handle = *com_handle_pointer;
    DWORD bytes_read;
    // LPOVERLAPPED OverlappedStruct;
    // Read data from the Serial Port
    BOOL status = ReadFile(com_handle,
            message,
            buffer_size,
            &bytes_read,
            NULL);
    
    if (!status)
    {
        printf("Serial read error\n");
        return SERIAL_READ_ERROR;
    }

    // If the number of bytes read is equal to the read size then there may be more bytes available in the buffer
    // I'm not sure this is the correct way to handle it but works for now
    if (bytes_read == buffer_size)
    {
        return SERIAL_BUFFER_NOT_FULLY_READ;
    }

    // Add a null to the end of the read string if it is done reading
    *(message + bytes_read) = '\0';
    
    return return_status;

}